VisualHeal = LibStub("AceAddon-3.0"):NewAddon("VisualHeal", "AceConsole-3.0");
VisualHeal.EventFrame = CreateFrame("Frame");
VisualHeal.EventFrame:SetScript("OnEvent", function (this, event, ...) VisualHeal[event](VisualHeal, ...) end);
VisualHeal.SharedMedia = LibStub("LibSharedMedia-3.0");

--[ Frequently Accessed Globals ]--

local UnitName = UnitName;
local UnitHealth = UnitHealth;
local UnitHealthMax = UnitHealthMax;
local UnitClass = UnitClass;
local UnitIsPlayer = UnitIsPlayer;
local RAID_CLASS_COLORS = RAID_CLASS_COLORS;
local select = select;

--[ Frame Factory and Frame Chains ]--

local incoming;
local freeList;
local outgoingList = {};
local outgoingTable = {};

function VisualHeal:LayoutBars()
    local position = self.db.profile.OutgoingPosition;
    local offset = 0;
    for i, frame in ipairs(outgoingList) do
        frame:SetPoint(position.point, position.parent, position.relpoint, position.x, position.y + offset);
        offset = offset + frame:GetHeight() + 5;
    end
end

function VisualHeal:CheckGUID(frame)
    local unitGUID = UnitGUID(frame.unit);
    if (unitGUID ~= frame.unitGUID) then
        if (UnitGUID("target") == frame.unitGUID) then
            frame.unit = "target";
        elseif (UnitGUID("focus") == frame.unitGUID) then
            frame.unit = "focus";
        else
            self:BarRelease(frame);
            return true;
        end
    end
    return false;
end

function VisualHeal:IncomingBarUpdate()
    local frame = incoming or self:BarCreate(true);
    incoming = frame;

    local width = frame:GetWidth();

    -- Update health
    local health = UnitHealth("player");
    local healthMax = UnitHealthMax("player");
    local hp = health / healthMax;
    frame.health:SetPoint("BOTTOMRIGHT", frame.backdrop, "BOTTOMLEFT", width * hp, 0);
    frame.health:SetVertexColor(hp < 0.5 and 1.0 or 2.0 * (1.0 - hp), hp > 0.5 and 0.8 or 1.6 * hp, 0.0);

    -- Update incoming heal
    local heal = UnitGetIncomingHeals("player") or 0;
    local playerHeal = UnitGetIncomingHeals("player", "player") or 0;
    local othersHeal = heal - playerHeal;
    if (othersHeal > 0) then
        frame.othersHeal:SetPoint("BOTTOMRIGHT", frame.health, "BOTTOMRIGHT", width * othersHeal / healthMax, 0);
        frame:Show();
    else
        frame:Hide();
        return;
    end
end

function VisualHeal:BarUpdate(frame)
    if self:CheckGUID(frame) then return end;

    local width = frame:GetWidth();

    -- Update health
    local health = UnitHealth(frame.unit);
    local healthMax = UnitHealthMax(frame.unit);
    local hp = health / healthMax;
    frame.health:SetPoint("BOTTOMRIGHT", frame.backdrop, "BOTTOMLEFT", width * hp, 0);
    frame.health:SetVertexColor(hp < 0.5 and 1.0 or 2.0 * (1.0 - hp), hp > 0.5 and 0.8 or 1.6 * hp, 0.0);

    -- Update player's heal
    local playerHeal = UnitGetIncomingHeals(frame.unit, "player");
    frame.playerHeal:SetPoint("BOTTOMRIGHT", frame.health, "BOTTOMRIGHT", width * playerHeal / healthMax, 0);
    local waste = 0;
    if (playerHeal > (healthMax - health)) then
        waste = (playerHeal - (healthMax - health)) / playerHeal;
    end
    local red = waste > 0.1 and 1 or waste * 10;
    local green = waste < 0.1 and 1 or -2.5 * waste + 1.25;
    if (waste < 0.0) then
        green = 1.0;
        red = 0.0;
    end
    frame.playerHeal:SetVertexColor(red, green, 0.0);

    -- Update other's heal
    local othersHeal = UnitGetIncomingHeals(frame.unit) - playerHeal;
    frame.othersHeal:SetPoint("BOTTOMRIGHT", frame.health, "RIGHT", width * othersHeal / healthMax, 0);
end

function VisualHeal:BarConfigure(frame)
    local barTexture = self.SharedMedia:Fetch('statusbar', self.db.profile.BarTexture);
    local borderTexture = self.SharedMedia:Fetch("border", self.db.profile.BorderTexture);
    local font = self.SharedMedia:Fetch("font", self.db.profile.Font);
    local width = 200;
    local height = 15;
    local edgeSize = (height - 2) / 1.5;
    local borderWidth = (edgeSize + 2 / 1.5) / 2;
    if (edgeSize > 16) then
        edgeSize = 16;
        borderWidth = edgeSize / 2;
    end
    frame.border:SetBackdrop({edgeFile = borderTexture, edgeSize = edgeSize});
    frame.border:SetBackdropBorderColor(self.db.profile.BorderBrightness, self.db.profile.BorderBrightness, self.db.profile.BorderBrightness, 1.0);
    frame.border:SetWidth(width + borderWidth - 2);
    frame.border:SetHeight(height + borderWidth - 2);

    frame.health:SetTexture(barTexture);

    if (not frame.isIncoming) then
        frame.playerHeal:SetTexture(barTexture);
        frame.playerHeal:SetVertexColor(0, 0, 1, 1);
    end

    frame.othersHeal:SetTexture(barTexture);
    frame.othersHeal:SetVertexColor(0.35, 0.55, 0.35, 0.6);

    frame.text:SetFont(font, self.db.profile.FontSize, self.db.profile.FontOutline);

    frame:SetWidth(width);
    frame:SetHeight(height);
    frame.spark:SetHeight(height * 1.5);

    if (frame.isIncoming) then
        local position = self.db.profile.IncomingPosition;
        frame:SetPoint(position.point, position.parent, position.relpoint, position.x, position.y);
    end
end

function VisualHeal:BarCreate(isIncoming)
    local frame = CreateFrame("Frame", nil, UIParent);
    frame:Hide();
    frame.isIncoming = isIncoming;
    frame:SetFrameStrata("HIGH");
    frame.border = CreateFrame("Frame", nil, frame);
    frame.border:SetPoint("CENTER", frame, "CENTER", 0, 0);
    frame.border:Show();

    frame.backdrop = frame:CreateTexture(nil, "BACKGROUND");
    frame.backdrop:SetPoint("TOPLEFT", frame ,"TOPLEFT", 0, 0)
    frame.backdrop:SetPoint("BOTTOMRIGHT", frame ,"BOTTOMRIGHT", 0, 0)
    frame.backdrop:SetTexture(0, 0, 0, 0.5)

    frame.health = frame:CreateTexture(nil, "BORDER")
    frame.health:SetPoint("TOPLEFT", frame.backdrop, "TOPLEFT", 0, 0)

    if (not isIncoming) then
        frame.playerHeal = frame:CreateTexture(nil, "BORDER")
        frame.playerHeal:ClearAllPoints()
        frame.playerHeal:SetPoint("TOPLEFT", frame.health, "TOPRIGHT", 0, 0)
    end

    frame.othersHeal = frame:CreateTexture(nil, "ARTWORK")
    frame.othersHeal:ClearAllPoints()
    frame.othersHeal:SetPoint("TOPLEFT", frame.health, "TOPRIGHT", 0, 0)

    frame.spark = frame:CreateTexture(nil, "OVERLAY");
    frame.spark:SetTexture("Interface\\CastingBar\\UI-CastingBar-Spark");
    frame.spark:SetVertexColor(1, 1, 1, 1)
    frame.spark:SetBlendMode('ADD');
    frame.spark:SetWidth(10);
    frame.spark:SetPoint("CENTER", frame.health, "RIGHT", 0, 0);

    frame.text = frame:CreateFontString(nil, "OVERLAY", "GameFontHighlight");
    frame.text:SetPoint("CENTER", frame.backdrop, "CENTER", 0, 0);

    self:BarConfigure(frame);
    return frame;
end

function VisualHeal:BarAllocate(unit, unitGUID)
    local frame;
    if (freeList) then
        frame = freeList;
        freeList = freeList.nextFrame;
    else
        frame = self:BarCreate();
    end
    frame.unit = unit;
    frame.unitGUID = unitGUID;

    -- Text
    frame.text:SetText(UnitName(unit));
    if (UnitIsPlayer(unit)) then
        local tab = RAID_CLASS_COLORS[select(2, UnitClass(unit)) or ""];
        frame.text:SetTextColor(tab.r, tab.g, tab.b);
    else
        frame.text:SetTextColor(230, 230, 0);
    end

    table.insert(outgoingList, frame);
    outgoingTable[unitGUID] = frame;
    frame:Show();
    self:LayoutBars();
    return frame;
end

function VisualHeal:BarRelease(frame)
    frame:Hide();
    frame.nextFrame = freeList;
    freeList = frame;
    outgoingTable[frame.unitGUID] = nil;
    for i, v in pairs(outgoingList) do
        if (v == frame) then
            table.remove(outgoingList, i);
            break;
        end
    end
    self:LayoutBars();
end

--[ Settings ]--

function VisualHeal:GetOptionsTable()
    local optionstable =
    {
        type = "group",
        icon = "",
        name = "VisualHeal",
        handler = self,
        childGroups = "tree",
        args =
        {
            intro = {
                order = 1,
                type = "description",
                name = "VisualHeal is a small and simple but powerful tool for all classes that visually shows your heals to others and heals incoming to you from others.",
            },
            general =
            {
                type = "group",
                name = "General Options",
                guiInline = true,
                order = 10,
                args =
                {
                    healbar =
                    {
                        name = "Enable HealBar",
                        type = "toggle",
                        order = 1,
                        desc = "Toggles display of the HealBar when you are healing",
                        get = function() return self.db.profile.ShowHealBar end,
                        set = function() self.db.profile.ShowHealBar = not self.db.profile.ShowHealBar end,
                    },
                    playerbar =
                    {
                        name = "Enable PlayerBar",
                        type = "toggle",
                        order = 2,
                        desc = "Toggles display of the PlayerBar when heals are incoming to you",
                        get = function() return self.db.profile.ShowPlayerBar end,
                        set = function() self.db.profile.ShowPlayerBar = not self.db.profile.ShowPlayerBar end,
                    },
                },
            },
            appearance =
            {
                type = "group",
                name = "Appearance Options",
                guiInline = true,
                order = 20,
                args =
                {
                    show =
                    {
                        name = "Show Bars",
                        type = "execute",
                        order = 1,
                        desc = "Show the HealBar and PlayerBar to allow moving them around",
                        func = "ShowBars",
                    },
                    texture =
                    {
                        name = "Bar Texture",
                        type = 'select',
                        order = 5,
                        desc = "Select texture to use for the HealBar and PlayerBar",
                        dialogControl = 'LSM30_Statusbar',
                        values = AceGUIWidgetLSMlists.statusbar,
                        get = function()
                            return self.db.profile.BarTexture;
                        end,
                        set = function(info, value)
                            self.db.profile.BarTexture = value;
                        end,
                    },
                    border =
                    {
                        name = "Border Texture",
                        type = 'select',
                        order = 6,
                        desc = "Select texture to use for the border of the HealBar and PlayerBar",
                        dialogControl = 'LSM30_Border',
                        values = AceGUIWidgetLSMlists.border,
                        get = function()
                            return self.db.profile.BorderTexture;
                        end,
                        set = function(info, value)
                            self.db.profile.BorderTexture = value;
                        end,
                    },
                    borderbrightness =
                    {
                        name = "Border Brightness",
                        type = "range",
                        order = 7,
                        desc = "Set the brightness of the border of the HealBar and PlayerBar",
                        min = 0.0,
                        max = 1.0,
                        isPercent = true,
                        get = function() return self.db.profile.BorderBrightness end,
                        set = function(info, value)
                            self.db.profile.BorderBrightness = value;
                        end,
                    },
                    font =
                    {
                        name = "Font",
                        type = "select",
                        order = 8,
                        desc = "Select font to use for the bars",
                        dialogControl = "LSM30_Font",
                        values = AceGUIWidgetLSMlists.font,
                        get = function() return self.db.profile.Font end,
                        set = function(info, value)
                            self.db.profile.Font = value;
                        end,
                    },
                    fontsize =
                    {
                        name = "Font Size",
                        type = "range",
                        order = 9,
                        desc = "Set the font size of the bars",
                        min = 6,
                        max = 30,
                        step = 1,
                        get = function() return self.db.profile.FontSize end,
                        set = function(info, value)
                            self.db.profile.FontSize = value;
                        end,
                    },
                    fontoutline =
                    {
                        name = "Font Outline",
                        type = "toggle",
                        order = 10,
                        desc = "Toggles outline on the font of the bars",
                        get = function() return self.db.profile.FontOutline end,
                        set = function()
                            self.db.profile.FontOutline = not self.db.profile.FontOutline;
                        end,
                    },
                },
            },
        },
    };
    return optionstable;
end

function VisualHeal:OpenConfig()
    InterfaceOptionsFrame_OpenToCategory(self.BlizzardOptionsFrame);
end

function VisualHeal:FrameMoved(frame)
    local layout = self.db.profile[frame:GetName() .. "Layout"];
    local point, parent, relpoint, x, y = frame:GetPoint();
    if (parent) then
        layout.default = true;
        layout.point, layout.relpoint, layout.x, layout.y = "CENTER", "CENTER", 0, 0;
    else
        layout.default = false;
        layout.point, layout.relpoint, layout.x, layout.y = point, relpoint, x, y;
    end
end

--[ ConfigMode Support ]--

CONFIGMODE_CALLBACKS = CONFIGMODE_CALLBACKS or {};
CONFIGMODE_CALLBACKS["VisualHeal"] =
function(action)
    if (action == "ON") then
        VisualHeal:ShowBars();
    elseif (action == "OFF") then
        VisualHeal:HideBars();
    end
end

--[ Init/Enable/Disable ]--

function VisualHeal:OnInitialize()
    self.db = LibStub("AceDB-3.0"):New("VisualHealDB",
    {
        profile =
        {
            ShowHealBar = true,
            ShowPlayerBar = true,
            BorderBrightness = 0.2,
            Font = "Friz Quadrata TT",
            FontSize = 9,
            FontOutline = true,
            OutgoingPosition = {point = "CENTER", relpoint = "CENTER", x = 0, y = -115},
            IncomingPosition = {point = "CENTER", relpoint = "CENTER", x = 0, y = -155},
        }
    });

    LibStub("AceConfig-3.0"):RegisterOptionsTable("VisualHeal", self:GetOptionsTable());
    self.BlizzardOptionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("VisualHeal", "VisualHeal");
    self:RegisterChatCommand("vh", "OpenConfig");
    self:RegisterChatCommand("visualheal", "OpenConfig");

    local playerClass = string.lower(select(2, UnitClass('player')));

    if (playerClass == "druid" or playerClass == "paladin" or playerClass == "priest" or playerClass == "shaman") then
        self.IsHealer = true;
    end
end

function VisualHeal:OnEnable()
    self.EventFrame:RegisterEvent("UNIT_HEALTH");
    self.EventFrame:RegisterEvent("UNIT_MAXHEALTH")
    self.EventFrame:RegisterEvent("PLAYER_TARGET_CHANGED")
    self.EventFrame:RegisterEvent("PLAYER_FOCUS_CHANGED")
    self.EventFrame:RegisterEvent("UNIT_HEAL_PREDICTION");
    self.SharedMedia.RegisterCallback(self, "LibSharedMedia_Registered", "LibSharedMedia_Callback")
    self.SharedMedia.RegisterCallback(self, "LibSharedMedia_SetGlobal", "LibSharedMedia_Callback")
end

function VisualHeal:OnDisable()
    self.EventFrame:UnregisterEvent("UNIT_HEALTH");
    self.EventFrame:UnregisterEvent("UNIT_MAXHEALTH")
    self.EventFrame:UnregisterEvent("PLAYER_TARGET_CHANGED")
    self.EventFrame:UnregisterEvent("PLAYER_FOCUS_CHANGED")
    self.EventFrame:UnregisterEvent("UNIT_HEAL_PREDICTION");
end

--[ Event Handlers ]--

function VisualHeal:UNIT_HEALTH(unit)
    if (unit == "player") then
        self:IncomingBarUpdate();
    end
    local unitGUID = UnitGUID(unit);
    local frame = outgoingTable[unitGUID];
    if (frame) then
        self:BarUpdate(frame);
    end
end

function VisualHeal:UNIT_MAXHEALTH(unit)
    self:UNIT_HEALTH(unit);
end

function VisualHeal:UnitUpdate(unit)
    local playerHeal = UnitGetIncomingHeals(unit, "player") or 0;
    local unitGUID = UnitGUID(unit);
    local frame = outgoingTable[unitGUID];
    if (playerHeal == 0 and frame) then
        self:BarRelease(frame);
    elseif (playerHeal > 0) then
        if (not frame) then
            frame = self:BarAllocate(unit, unitGUID);
        end
        self:BarUpdate(frame);
    end
end

local lastTargetGUID;
function VisualHeal:PLAYER_TARGET_CHANGED()
    local frame = outgoingTable[lastTargetGUID];
    if (lastTargetGUID and frame) then
        self:CheckGUID(frame);
    end
    lastTargetGUID = UnitGUID("target");
    self:UnitUpdate("target");
end

local lastFocusGUID;
function VisualHeal:PLAYER_FOCUS_CHANGED()
    local frame = outgoingTable[lastFocusGUID];
    if (lastFocusGUID and frame) then
        self:CheckGUID(frame);
    end
    lastFocusGUID = UnitGUID("focus");
    self:UnitUpdate("focus");
end

function VisualHeal:UNIT_HEAL_PREDICTION(unit)
    if (unit == "player") then
        self:IncomingBarUpdate();
    end
    self:UnitUpdate(unit);
end

function VisualHeal:LibSharedMedia_Callback()
    if (incoming) then
        self:BarConfigure(incoming);
    end
    for i, frame in ipairs(outgoingList) do
        self:BarConfigure(frame);
    end
    local frame = freeList;
    while (frame) do
        self:BarConfigure(frame);
        frame = frame.nextFrame;
    end
end
